@@ -1,6 +1,5 @@
 #
 #               tk/canvas.rb - Tk canvas classes
-#                       $Date: 2005/10/24 00:07:00 $
 #                       by Yukihiro Matsumoto <matz@caelum.co.jp>
 #
 require 'tk'
@@ -14,7 +13,7 @@ module TkCanvasItemConfig
   def __item_strval_optkeys(id)
     # maybe need to override
     super(id) + [
-      'fill', 'activefill', 'disabledfill', 
+      'fill', 'activefill', 'disabledfill',
       'outline', 'activeoutline', 'disabledoutline'
     ]
   end
@@ -26,7 +25,10 @@ module TkCanvasItemConfig
   private :__item_methodcall_optkeys
 
   def __item_val2ruby_optkeys(id)  # { key=>proc, ... }
-    super(id).update('window'=>proc{|i, v| window(v)})
+    super(id).update('window'=>proc{|i, v| window(v)},
+                     'tags'=>proc{|i, v|
+                       simplelist(v).collect{|tag| TkcTag.id2obj(self, tag)}
+                     })
   end
   private :__item_val2ruby_optkeys
 
@@ -40,13 +42,13 @@ module TkCanvasItemConfig
   private :__item_pathname
 end
 
-class TkCanvas<TkWindow
+class Tk::Canvas<TkWindow
   include TkCanvasItemConfig
-  include Scrollable
+  include Tk::Scrollable
 
   TkCommandNames = ['canvas'.freeze].freeze
   WidgetClassName = 'Canvas'.freeze
-  WidgetClassNames[WidgetClassName] = self
+  WidgetClassNames[WidgetClassName] ||= self
 
   def __destroy_hook__
     TkcItem::CItemID_TBL.delete(@path)
@@ -83,6 +85,7 @@ class TkCanvas<TkWindow
 
   # create a canvas item without creating a TkcItem object
   def create(type, *args)
+    type = TkcItem.type2class(type.to_s) unless type.kind_of?(TkcItem)
     type.create(self, *args)
   end
 
@@ -118,7 +121,7 @@ class TkCanvas<TkWindow
   end
 
   def bbox(tagOrId, *tags)
-    list(tk_send_without_enc('bbox', tagid(tagOrId), 
+    list(tk_send_without_enc('bbox', tagid(tagOrId),
                              *tags.collect{|t| tagid(t)}))
   end
 
@@ -169,9 +172,11 @@ class TkCanvas<TkWindow
     #tk_tcl2ruby(tk_send_without_enc('canvasy', screen_y, *args))
     number(tk_send_without_enc('canvasy', screen_y, *args))
   end
+  alias canvas_x canvasx
+  alias canvas_y canvasy
 
   def coords(tag, *args)
-    if args == []
+    if args.empty?
       tk_split_list(tk_send_without_enc('coords', tagid(tag)))
     else
       tk_send_without_enc('coords', tagid(tag), *(args.flatten))
@@ -180,17 +185,23 @@ class TkCanvas<TkWindow
   end
 
   def dchars(tag, first, last=None)
-    tk_send_without_enc('dchars', tagid(tag), 
+    tk_send_without_enc('dchars', tagid(tag),
                         _get_eval_enc_str(first), _get_eval_enc_str(last))
     self
   end
 
   def delete(*args)
-    if TkcItem::CItemID_TBL[self.path]
+    tbl = nil
+    TkcItem::CItemID_TBL.mutex.synchronize{
+      tbl = TkcItem::CItemID_TBL[self.path]
+    }
+    if tbl
       args.each{|tag|
         find('withtag', tag).each{|item|
           if item.kind_of?(TkcItem)
-            TkcItem::CItemID_TBL[self.path].delete(item.id)
+            TkcItem::CItemID_TBL.mutex.synchronize{
+              tbl.delete(item.id)
+            }
           end
         }
       }
@@ -207,7 +218,7 @@ class TkCanvas<TkWindow
   alias deltag dtag
 
   def find(mode, *args)
-    list(tk_send_without_enc('find', mode, *args)).collect!{|id| 
+    list(tk_send_without_enc('find', mode, *args)).collect!{|id|
       TkcItem.id2obj(self, id)
     }
   end
@@ -258,12 +269,18 @@ class TkCanvas<TkWindow
     self
   end
 
+  def imove(tagOrId, idx, x, y)
+    tk_send_without_enc('imove', tagid(tagOrId), idx, x, y)
+    self
+  end
+  alias i_move imove
+
   def index(tagOrId, idx)
     number(tk_send_without_enc('index', tagid(tagOrId), idx))
   end
 
   def insert(tagOrId, index, string)
-    tk_send_without_enc('insert', tagid(tagOrId), index, 
+    tk_send_without_enc('insert', tagid(tagOrId), index,
                         _get_eval_enc_str(string))
     self
   end
@@ -293,7 +310,7 @@ class TkCanvas<TkWindow
         fnt
       end
     else
-      tk_tcl2ruby(_fromUTF8(tk_send_without_enc('itemcget', tagid(tagOrId), 
+      tk_tcl2ruby(_fromUTF8(tk_send_without_enc('itemcget', tagid(tagOrId),
                                                 "-#{option}")))
     end
   end
@@ -308,16 +325,16 @@ class TkCanvas<TkWindow
           || key['latinfont'] || key['asciifont'] )
         tagfont_configure(tagid(tagOrId), key.dup)
       else
-        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId), 
+        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId),
                                       *hash_kv(key, true)))
       end
 
     else
       if ( key == 'coords' || key == :coords )
         self.coords(tagOrId, value)
-      elsif ( key == 'font' || key == :font || 
-              key == 'kanjifont' || key == :kanjifont || 
-              key == 'latinfont' || key == :latinfont || 
+      elsif ( key == 'font' || key == :font ||
+              key == 'kanjifont' || key == :kanjifont ||
+              key == 'latinfont' || key == :latinfont ||
               key == 'asciifont' || key == :asciifont )
         if value == None
           tagfontobj(tagid(tagOrId))
@@ -325,7 +342,7 @@ class TkCanvas<TkWindow
           tagfont_configure(tagid(tagOrId), {key=>value})
         end
       else
-        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId), 
+        _fromUTF8(tk_send_without_enc('itemconfigure', tagid(tagOrId),
                                       "-#{key}", _get_eval_enc_str(value)))
       end
     end
@@ -382,16 +399,16 @@ class TkCanvas<TkWindow
           else
             if conf[3]
               if conf[3].index('{')
-                conf[3] = tk_split_list(conf[3]) 
+                conf[3] = tk_split_list(conf[3])
               else
-                conf[3] = tk_tcl2ruby(conf[3]) 
+                conf[3] = tk_tcl2ruby(conf[3])
               end
             end
             if conf[4]
               if conf[4].index('{')
-                conf[4] = tk_split_list(conf[4]) 
+                conf[4] = tk_split_list(conf[4])
               else
-                conf[4] = tk_tcl2ruby(conf[4]) 
+                conf[4] = tk_tcl2ruby(conf[4])
               end
             end
           end
@@ -414,8 +431,8 @@ class TkCanvas<TkWindow
         when 'coords'
           {'coords' => ['', '', '', self.coords(tagOrId)]}
         when 'dash', 'activedash', 'disableddash'
-          conf = tk_split_simplelist(tk_send_without_enc('itemconfigure', 
-                                                         tagid(tagOrId), 
+          conf = tk_split_simplelist(tk_send_without_enc('itemconfigure',
+                                                         tagid(tagOrId),
                                                          "-#{key}"))
           if conf[3] && conf[3] =~ /^[0-9]/
             conf[3] = list(conf[3])
@@ -450,16 +467,16 @@ class TkCanvas<TkWindow
           else
             if conf[2]
               if conf[2].index('{')
-                conf[2] = tk_split_list(conf[2]) 
+                conf[2] = tk_split_list(conf[2])
               else
-                conf[2] = tk_tcl2ruby(conf[2]) 
+                conf[2] = tk_tcl2ruby(conf[2])
               end
             end
             if conf[3]
               if conf[3].index('{')
-                conf[3] = tk_split_list(conf[3]) 
+                conf[3] = tk_split_list(conf[3])
               else
-                conf[3] = tk_tcl2ruby(conf[3]) 
+                conf[3] = tk_tcl2ruby(conf[3])
               end
             end
           end
@@ -516,11 +533,18 @@ class TkCanvas<TkWindow
     self
   end
 
-  def move(tag, x, y)
-    tk_send_without_enc('move', tagid(tag), x, y)
+  def move(tag, dx, dy)
+    tk_send_without_enc('move', tagid(tag), dx, dy)
     self
   end
 
+  def moveto(tag, x, y)
+    # Tcl/Tk 8.6 or later
+    tk_send_without_enc('moveto', tagid(tag), x, y)
+    self
+  end
+  alias move_to moveto
+
   def postscript(keys)
     tk_send("postscript", *hash_kv(keys))
   end
@@ -534,6 +558,15 @@ class TkCanvas<TkWindow
     self
   end
 
+  def rchars(tag, first, last, str_or_coords)
+    # Tcl/Tk 8.6 or later
+    str_or_coords = str_or_coords.flatten if str_or_coords.kinad_of? Array
+    tk_send_without_enc('rchars', tagid(tag), first, last, str_or_coords)
+    self
+  end
+  alias replace_chars rchars
+  alias replace_coords rchars
+
   def scale(tag, x, y, xs, ys)
     tk_send_without_enc('scale', tagid(tag), x, y, xs, ys)
     self
@@ -543,8 +576,8 @@ class TkCanvas<TkWindow
     tk_send_without_enc('scan', 'mark', x, y)
     self
   end
-  def scan_dragto(x, y)
-    tk_send_without_enc('scan', 'dragto', x, y)
+  def scan_dragto(x, y, gain=None)
+    tk_send_without_enc('scan', 'dragto', x, y, gain)
     self
   end
 
@@ -571,8 +604,37 @@ class TkCanvas<TkWindow
   def itemtype(tag)
     TkcItem.type2class(tk_send('type', tagid(tag)))
   end
+
+  def create_itemobj_from_id(idnum)
+    id = TkcItem.id2obj(self, idnum.to_i)
+    return id if id.kind_of?(TkcItem)
+
+    typename = tk_send('type', id)
+    unless type = TkcItem.type2class(typename)
+      (itemclass = typename.dup)[0,1] = typename[0,1].upcase
+      type = TkcItem.const_set(itemclass, Class.new(TkcItem))
+      type.const_set("CItemTypeName", typename.freeze)
+      TkcItem::CItemTypeToClass[typename] = type
+    end
+
+    canvas = self
+    (obj = type.allocate).instance_eval{
+      @parent = @c = canvas
+      @path = canvas.path
+      @id = id
+      TkcItem::CItemID_TBL.mutex.synchronize{
+        TkcItem::CItemID_TBL[@path] = {} unless TkcItem::CItemID_TBL[@path]
+        TkcItem::CItemID_TBL[@path][@id] = self
+      }
+    }
+  end
 end
 
+#TkCanvas = Tk::Canvas unless Object.const_defined? :TkCanvas
+#Tk.__set_toplevel_aliases__(:Tk, Tk::Canvas, :TkCanvas)
+Tk.__set_loaded_toplevel_aliases__('tk/canvas.rb', :Tk, Tk::Canvas, :TkCanvas)
+
+
 class TkcItem<TkObject
   extend Tk
   include TkcTagAccess
@@ -581,9 +643,12 @@ class TkcItem<TkObject
 
   CItemTypeName = nil
   CItemTypeToClass = {}
+
   CItemID_TBL = TkCore::INTERP.create_table
 
-  TkCore::INTERP.init_ip_env{ CItemID_TBL.clear }
+  TkCore::INTERP.init_ip_env{
+    CItemID_TBL.mutex.synchronize{ CItemID_TBL.clear }
+  }
 
   def TkcItem.type2class(type)
     CItemTypeToClass[type]
@@ -591,8 +656,13 @@ class TkcItem<TkObject
 
   def TkcItem.id2obj(canvas, id)
     cpath = canvas.path
-    return id unless CItemID_TBL[cpath]
-    CItemID_TBL[cpath][id]? CItemID_TBL[cpath][id]: id
+    CItemID_TBL.mutex.synchronize{
+      if CItemID_TBL[cpath]
+        CItemID_TBL[cpath][id]? CItemID_TBL[cpath][id]: id
+      else
+        id
+      end
+    }
   end
 
   ########################################
@@ -625,6 +695,13 @@ class TkcItem<TkObject
         fontkeys[fkey] = keys.delete(fkey) if keys.key?(fkey)
       }
 
+      __item_optkey_aliases(nil).each{|alias_name, real_name|
+        alias_name = alias_name.to_s
+        if keys.has_key?(alias_name)
+          keys[real_name.to_s] = keys.delete(alias_name)
+        end
+      }
+
       __item_methodcall_optkeys(nil).each{|key|
         key = key.to_s
         methodkeys[key] = keys.delete(key) if keys.key?(key)
@@ -641,7 +718,7 @@ class TkcItem<TkObject
       args = args.flatten
     end
 
-    [args, fontkeys]
+    [args, fontkeys, methodkeys]
   end
   private_class_method :_parse_create_args
 
@@ -649,24 +726,27 @@ class TkcItem<TkObject
     unless self::CItemTypeName
       fail RuntimeError, "#{self} is an abstract class"
     end
-    args, fontkeys = _parse_create_args(args)
-    idnum = tk_call_without_enc(canvas.path, 'create', 
+    args, fontkeys, methodkeys = _parse_create_args(args)
+    idnum = tk_call_without_enc(canvas.path, 'create',
                                 self::CItemTypeName, *args)
     canvas.itemconfigure(idnum, fontkeys) unless fontkeys.empty?
+    canvas.itemconfigure(idnum, methodkeys) unless methodkeys.empty?
     idnum.to_i  # 'canvas item id' is an integer number
   end
   ########################################
 
   def initialize(parent, *args)
-    #unless parent.kind_of?(TkCanvas)
-    #  fail ArgumentError, "expect TkCanvas for 1st argument"
+    #unless parent.kind_of?(Tk::Canvas)
+    #  fail ArgumentError, "expect Tk::Canvas for 1st argument"
     #end
     @parent = @c = parent
     @path = parent.path
 
     @id = create_self(*args) # an integer number as 'canvas item id'
-    CItemID_TBL[@path] = {} unless CItemID_TBL[@path]
-    CItemID_TBL[@path][@id] = self
+    CItemID_TBL.mutex.synchronize{
+      CItemID_TBL[@path] = {} unless CItemID_TBL[@path]
+      CItemID_TBL[@path][@id] = self
+    }
   end
   def create_self(*args)
     self.class.create(@c, *args) # return an integer number as 'canvas item id'
@@ -687,7 +767,9 @@ class TkcItem<TkObject
 
   def delete
     @c.delete @id
-    CItemID_TBL[@path].delete(@id) if CItemID_TBL[@path]
+    CItemID_TBL.mutex.synchronize{
+      CItemID_TBL[@path].delete(@id) if CItemID_TBL[@path]
+    }
     self
   end
   alias remove  delete
